/* dispint.h
 *
 * Asp Dispatch Interface definition
 *
 * REVISIONS
 * 	1.0		10-1-98		Dave Weaver
 *	- created
 *	1.1		6-3-99		Dave Weaver
 *	- added support for virtual hosts
 *	1.2		9-9-99		Dave Weaver
 *	- added loaded api
 *
 */
 
#ifndef __DISPINT_H_
#define __DISPINT_H_

#ifdef __cplusplus
extern "C" {
#endif

#ifndef DWORD
typedef unsigned long DWORD;
typedef unsigned long * LPDWORD;
#endif

#ifndef STDCALL
#ifdef WIN32
#define STDCALL	__stdcall
#else
#define STDCALL
#define MAKELONG(a,b) ((long)(((unsigned short)(a)) | ((unsigned long)((unsigned short)(b))) << 16))
#endif 
#endif 

#define HT_VERSION_MAJOR			1
#define HT_VERSION_MINOR			2
#define HT_VERSION				MAKELONG(HT_VERSION_MINOR,HT_VERSION_MAJOR)

#define HT_TERM_ADVISORY_UNLOAD		0x00000001
#define HT_TERM_MUST_UNLOAD			0x00000002

#define HT_STATUS_SUCCESS			0x00000000
#define HT_STATUS_ERROR				0x00000001

typedef void*  HCONN;

#define HT_SFUNC_LOGDATA			0x00000001
#define HT_SFUNC_RESPONSEHEADERS	0x00000002

#define HT_MAX_CONTENT_LENGTH		49152		/* 48k */

/* init function flags */
#define UPDATE						0x00000001
#define USEDEFAULTS					0x00000002

/* identifier for real host */
#define HT_REALHOST_ID "/"

/* macro to determine if host is real or virtual */
#define HT_IS_VIRTUALHOST(phrb) (phrb && phrb->lpszVirtualHost && phrb->lpszVirtualHost[0] != '/')?1:0

/* passed to http_server_load */
typedef struct _HT_LOAD_INFO
{
	DWORD dwVersion;				/* version info of this spec */
	DWORD dwFlags;					/* flags used for the init function call */
	int bController;				/* is this the controlling engine ? */
} HT_LOAD_INFO, *LPHT_LOAD_INFO;

/* passed to http_server_init() */
typedef struct _HT_INIT_INFO 
{
	/* required for each call */

	DWORD dwVersion;				/* version info of this spec */
	DWORD dwFlags;					/* flags used for the init function call */
	int	nEngineid;					/* id for the engine to use */
	int	bEnabled;					/* enable/disable script execution */

	/* all the following can only be used at initialization, ignored otherwise */

	char szLibBinDir[1024];			/* path to engine lib directory */
	int	bMultiThreadedEngine;		/* is the asp engines multithreaded */
	int bMultiThreadedClient;		/* is the client multithreaded */
	char *lpszVirtualRoots;			/* double NULL terminated string of NULL delimited vroot/filepath pairs */
	DWORD cbVroots;					/* size of the vroot buffer */ 

	/* if USEDEFAULTS is specified, the following settings are ignored */

	int	nScriptTimeout;				/* default timeout for a running script */
	int	nSessionTimeout;			/* default session timeout */
	int	bBufferingOn;				/* buffering on/off by default */
	char szDefLanguage[50];			/* default scripting language */
	char szDefAccount[100];			/* default login account */
	char szDefPasswd[100];			/* default login password */
	char szDefDomain[100];			/* default login domain */
	unsigned int uDefTransaction;	/* default transaction mode */
	unsigned int uDefCodePage;		/* deafult codepage */
	long lDefLCID;					/* default LCID */
	int	nLogErrors;					/* should we log errors */
	char szLogDirectory[500];		/* directory where we log errors */
	char szDefError[5000];			/* default error message to report */
	int	bShowDefaultError;			/* show the default error message */
	int	bOutputDebug;				/* have engine spit out debug messages */
	int	bLogToFile;					/* should we log to debug messages to a file */
	int	bUseSessions;				/* allow sessions to be used */
	int	bAllowOutOfProcCmpnts;		/* allow out of process components to be used */
	int	nNumInitialThreads;			/* number of initial threads in the thread pool */
	int	nProcessorThreadMax;		/* max number of threads in thread pool */
	int	nThreadCreationThreshold;	/* how many waiting requests before adding a thread */

} HT_INIT_INFO, *LPHT_INIT_INFO;

/* passed to http_server_request() */
typedef struct _HT_REQUEST_BLOCK 
{
	DWORD cbSize;					/* size of this struct. */
	HCONN hConn;					/* Context number not to be modified! */

	DWORD dwVersion;				/* version info of this spec */

	char *lpszUserInfo;				/* user login information */
	char *lpszMethod;				/* REQUEST_METHOD */
	char *lpszQueryString;			/* QUERY_STRING */
	char *lpszPathInfo;				/* PATH_INFO */
	char *lpszPathTranslated;		/* PATH_TRANSLATED */
	char *lpszCookie;				/* HTTP_COOKIE */
	char *lpszContentType;			/* CONTENT_TYPE */

	void *lpData;					/* Pointer to available bytes */
	DWORD cbTotalBytes;				/* Total bytes indicated from client */
	DWORD cbAvailable;				/* Available number of bytes */

	char *lpszVirtualHost;			/* Virtual host id */
	char *lpszVirtualHostRoot;		/* Document root of virtual host */
	
	DWORD dwReserved;				/* reserved, set to zero */

	int (STDCALL *get_server_var)(HCONN hConn,
                                  char *lpszVariableName,
                                  void *lpvBuffer,
                                  LPDWORD lpluSize);

	int (STDCALL *write_client)(HCONN hConn,
                                void *Buffer,
                                LPDWORD lpluBytes);

	int (STDCALL *read_client)(HCONN hConn,
                               void *lpvBuffer,
                               LPDWORD lpluSize);

	int (STDCALL *server_function)(HCONN hConn,
                                   DWORD luHSERRequest,
                                   char *lpszBuffer,
                                   LPDWORD lpluSize,
								   char *lpszExtra);

} HT_REQUEST_BLOCK, *LPHT_REQUEST_BLOCK;


/* server functions */
DWORD STDCALL http_server_load(HT_LOAD_INFO *pHLI);
DWORD STDCALL http_server_init(HT_INIT_INFO *pHII);
DWORD STDCALL http_server_request(HT_REQUEST_BLOCK *pHRB, LPDWORD pdwStatus);
DWORD STDCALL http_server_term(void);

/* callbacks defined */
int STDCALL get_server_var(HCONN, char *, void *, LPDWORD);
int STDCALL write_client(HCONN, void *Buffer, LPDWORD); 
int STDCALL read_client(HCONN, void *lpvBuffer, LPDWORD);
int STDCALL server_function(HCONN, DWORD, char *, LPDWORD, char *);


#ifdef __cplusplus
}
#endif

#endif /* __DISPINT_H_ */
