/*
 * mod_casp2.c												
 *															
 * Apache module for chiliasp								
 * Dave Weaver - 10/1/98    								
 *															
 * Chili!Soft ASP Apache Module source code   				
 * Copyright (C) 1999 Chili!Soft Corporation          		
 * All rights reserved.                                     
 *															
 *															
 * RELEASE VERSION: 3.0.03	
 * STUB VERSION:	31								
 *
 *
 */

#ifndef WIN32
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_protocol.h"
#include "http_request.h"
#include "http_main.h"
#include "http_log.h"
#include "util_script.h"
#else /* WIN32 */
#include "ntapinc.h"
#endif /* !WIN32 */

/* chiliasp dispatcher interface 	*/
#include "dispint.h"	

/* setup version number and include headers */
/* for backwards compatability				*/
#if APACHE_RELEASE > 1030000
#define VER_13
#ifndef WIN32
#include "ap_compat.h" 
#endif /* WIN32  */
#else
#define VER_12 
#endif

int g_DebugFlag = 0;
#define error printf
#define	trace if (g_DebugFlag > 4) printf
#define	trace1 if (g_DebugFlag > 0) printf
#define	trace2 if (g_DebugFlag > 1) printf
#define	trace3 if (g_DebugFlag > 2) printf
#define	trace4 if (g_DebugFlag > 3) printf
#define	trace5 if (g_DebugFlag > 4) printf

#ifndef MODULE_VAR_EXPORT	/* only in version 1.3 */
#define MODULE_VAR_EXPORT 
#endif


/* macros for loading shared libraries */

#ifdef WIN32
#define DLHANDLE HINSTANCE
#define DLOPEN(l) LoadLibraryEx(l, NULL, LOAD_WITH_ALTERED_SEARCH_PATH)
#define DLSYM(h, s) GetProcAddress(h, s)
#define DLCLOSE(h) CloseHandle(h)
#else
#if defined( _AIX ) && defined(VER_13)
  /* Apache 1.3.x has its own dlopen, dlsym, etc routines */
  #include <../../os/unix/os.h>
#else
  #include <dlfcn.h>
#endif

typedef void *HINSTANCE;
#define DLHANDLE HINSTANCE

#ifndef _AIX
  #define DLOPEN(l) dlopen(l, RTLD_LAZY)
#else
  #define DLOPEN(l) dlopen(l, RTLD_GLOBAL|RTLD_NOW)
#endif /* _AIX */

#define DLSYM(h, s) dlsym(h, s)
#define DLCLOSE(h) dlclose(h)
#define strnicmp strncasecmp
#endif

#ifdef WIN32
#define STDCALL __stdcall
#else	// !WIN32
#define STDCALL
#endif	// !WIN32

typedef int (STDCALL *INITIALIZE_CASP_LIBS)(const char *, int);
typedef int (STDCALL *RELEASE_CASP_LIBS)();
typedef int (STDCALL *IS_CASP_LIBS_INITIALIZED)();
typedef int (STDCALL *SEARCH_APPLICATIONS)(const char *, const char *, char **);
typedef void (STDCALL *DELETE_APPLICATIONS)();
typedef void (STDCALL *ADD_APPLICATION)(const char *, const char *, const char *, int );
typedef DWORD (STDCALL *CASP_REQ)(LPHT_REQUEST_BLOCK phrb, LPDWORD pdw);
typedef int (STDCALL *IS_VIRTUAL_HOST_ALLOWED)(const char *);

static HINSTANCE hCasp2ap = NULL;
static char *real_host_name = NULL;
static INITIALIZE_CASP_LIBS initialize_casp_libs = NULL;
static RELEASE_CASP_LIBS release_casp_libs = NULL;
static IS_CASP_LIBS_INITIALIZED is_casp_libs_initialized = NULL;
static SEARCH_APPLICATIONS search_applications = NULL;
static DELETE_APPLICATIONS delete_applications = NULL;
static ADD_APPLICATION add_application = NULL;
static CASP_REQ casp_req = NULL;
static IS_VIRTUAL_HOST_ALLOWED is_virtual_host_allowed = NULL;


/* returns the servers unique id from the config data */
const char *get_server_id(request_rec *r);

/* returns the servers docroot from the config data */
const char *get_server_docroot(request_rec *r);

/* returns the servers user/group string from the config data */
const char *get_usergroup_string(request_rec *r);

/* casp module definition */
module MODULE_VAR_EXPORT casp2_module;  

/* chiliasp lib directory */
static char CaspLibDir[1024];

/* Our "Connection ID" structure */
typedef struct {
	LPHT_REQUEST_BLOCK lphrb;
	request_rec *r;
	int status;
	int custom_response;
} casp_cid;

/* 
 *  forward declarations
 */
int STDCALL get_server_var(HCONN hConn, char *lpszVariableName,
		   void *lpvBuffer, unsigned long *lpdwSizeofBuffer);
int STDCALL write_client(HCONN hConn, void *Buffer, unsigned long *lpwdwBytes); 
int STDCALL read_client(HCONN hConn, void *lpvBuffer, unsigned long *lpdwSize);
int STDCALL write_client(HCONN hConn, void *Buffer, unsigned long *lpwdwBytes);
int STDCALL server_function(HCONN hConn, unsigned long dwHSERequest,
		char *lpszBuffer, unsigned long *lpdwSize, char *lpszExtra);

#define OK 0
#define FAIL 1

/* load our helper library (located in [casproot]/lib */
static void init_casp2_module()
{
	/* only load it the first time */
	if (hCasp2ap == NULL)
	{
#ifdef WIN32
		hCasp2ap = DLOPEN("caspmod2ap.dll");
#else
		hCasp2ap = DLOPEN("libcasp2ap.so");
#endif
	
		if (!hCasp2ap)
		{
#ifdef _AIX
			fprintf(stderr, "%s\n", dlerror());
			fprintf(stderr, "chiliasp: Failed to open libcasp2ap.so."
				"\nCheck your LIBPATH\n environment variable.\n");
#elif WIN32
			error( "chiliasp: Failed to open caspmod2ap.dll.");
#else
			fprintf(stderr, "chiliasp: Failed to open libcasp2ap.so."
				"\nCheck your LD_LIBRARY_PATH environment variable.\n");
#endif /* _AIX */
			
			exit(1);
		}

		initialize_casp_libs = (INITIALIZE_CASP_LIBS) 
			DLSYM(hCasp2ap, "initialize_casp_libs");
		release_casp_libs = (RELEASE_CASP_LIBS)
			DLSYM(hCasp2ap, "release_casp_libs");
		is_casp_libs_initialized = (IS_CASP_LIBS_INITIALIZED)
			DLSYM(hCasp2ap, "is_casp_libs_initialized");
		search_applications = (SEARCH_APPLICATIONS)
			DLSYM(hCasp2ap, "search_applications");
		delete_applications = (DELETE_APPLICATIONS)
			DLSYM(hCasp2ap, "delete_applications");
		add_application = (ADD_APPLICATION)
			DLSYM(hCasp2ap, "add_application");
		casp_req = (CASP_REQ)
			DLSYM(hCasp2ap, "http_server_request");
		is_virtual_host_allowed = (IS_VIRTUAL_HOST_ALLOWED)
			DLSYM(hCasp2ap, "is_virtual_host_allowed");
	
		if (!initialize_casp_libs || !release_casp_libs || 
			!is_casp_libs_initialized || !search_applications ||
			!delete_applications || !add_application ||
			!is_virtual_host_allowed || !casp_req )
		{
			DLCLOSE(hCasp2ap);
#ifdef WIN32
			error("chiliasp: Failed to find required entry points points.\n");
#else
			fprintf(stderr, "chiliasp: Failed to find required entry points points.\n");
#endif
			exit(1);
		}
	}
}

void exit_handler(server_rec *s, pool *p)
{
	release_casp_libs(1);	
}

/* we are going to treat user roots (ie. /~davew/) as     */
/* virtual hosts                                          */
int translate_username_to_vhost(request_rec *r, HT_REQUEST_BLOCK *Hrb)
{
	if (!Hrb || !r)
		return FAIL;

	/* if it begins with a /~, then we assume it is a username */
    if (Hrb->lpszPathInfo && Hrb->lpszPathTranslated
        && Hrb->lpszPathInfo[0] == '/'  && Hrb->lpszPathInfo[1] == '~')
    {
        char *pos = strchr(Hrb->lpszPathInfo + 2, '/');
        if (pos)
        {
        	char *temp_virtual_host = NULL;
        	char *temp_virtual_host_root = NULL;
        	
        	/* allocate enoguh room for the new vhost name */
        	temp_virtual_host = pcalloc(r->pool, pos - Hrb->lpszPathInfo);
        	
            strncpy(temp_virtual_host, Hrb->lpszPathInfo + 2, pos - Hrb->lpszPathInfo - 2);
            temp_virtual_host[pos - Hrb->lpszPathInfo - 2] = 0;

            /* assume name is somewhere in the translated path */
            /* and strip off after it to create vhost root */
            pos = strstr(Hrb->lpszPathTranslated, pos);
            if (!pos)
            {
                fprintf(stderr, "chiliasp: pathinfo not found in translated path (%s)\n", Hrb->lpszPathInfo);
                return SERVER_ERROR;
            }
            
            /* now allocate for vhost root */
            if (get_server_docroot(r) == NULL)
            {
            	/* access is not allowed */
            	temp_virtual_host_root = NULL;
            }
            else
            {
	            temp_virtual_host_root = pcalloc(r->pool, strlen(Hrb->lpszPathTranslated) + 1);
	            strcpy(temp_virtual_host_root, Hrb->lpszPathTranslated);
	            temp_virtual_host_root[pos - Hrb->lpszPathTranslated] = 0;
            }
            
            Hrb->lpszVirtualHost = temp_virtual_host;
            Hrb->lpszVirtualHostRoot = temp_virtual_host_root;
            
            return OK;
        }
    }
    
    return FAIL;
}

#ifdef WIN32
/* Disable optimizer for this function on NT */
#pragma optimize("y",off) 
#endif /* WIN32 */

int casp_handler (request_rec *r) 
{
	HT_REQUEST_BLOCK *lphrb = pcalloc(r->pool, sizeof(struct _HT_REQUEST_BLOCK));
	casp_cid *cid = pcalloc(r->pool, sizeof(casp_cid));
	table *e = r->subprocess_env;
	int retval;
	unsigned long luStatus = 0;

	/* have we been initialized yet? */
	if (!is_casp_libs_initialized())
	{
		error("chiliasp: module not initialized.\n");
		return SERVER_ERROR;
	}
		
	/* Set up variables */
	add_common_vars(r);
	add_cgi_vars(r);
	
	/* Set up connection ID  */
	lphrb->hConn = (HCONN)cid;
	cid->lphrb = lphrb;
	cid->r = r;
	cid->status = 0;
	cid->custom_response = 0;

	/* set common http variables */
	lphrb->cbSize = sizeof(struct _HT_REQUEST_BLOCK);
	lphrb->dwVersion = HT_VERSION;
	lphrb->lpszMethod = (char *)r->method;
	lphrb->lpszQueryString = (char *)table_get(e, "QUERY_STRING");
	lphrb->lpszPathInfo = (char *)table_get(e, "SCRIPT_NAME");
	lphrb->lpszPathTranslated = (char *)r->filename;
	lphrb->lpszCookie = (char *)table_get(e, "HTTP_COOKIE");
	lphrb->lpszContentType = (char *)table_get(e, "CONTENT_TYPE");
	lphrb->dwReserved = 0;
	
	/* send the user and group ids that the 		*/
	/* chiliasp server will impersonate. this 		*/
	/* string has the format "userid/groupid"		*/
	lphrb->lpszUserInfo = (char *)get_usergroup_string(r);
	
    /* we are going to treat user roots (ie. /~davew/) as   */
    /* virtual hosts                                        */
    if (translate_username_to_vhost(r, lphrb) != OK)
    {
		/* send host name								*/
		/* if this is not the virtual host, then		*/
		/* send HT_REALHOST_ID							*/
		lphrb->lpszVirtualHost = (char *)get_server_id(r);
		
		/* send document root of host					*/
		/* if NULL is passed, chiliasp server 			*/
		/* not permit asp pages to be processed.		*/
		lphrb->lpszVirtualHostRoot = (char *)get_server_docroot(r);
	}
	
	/* cannot continue without a vhost identifier 	*/
	if (!lphrb->lpszVirtualHost)
		return SERVER_ERROR;

	trace1("chiliasp: PathInfo = %s\n", lphrb->lpszPathInfo?lphrb->lpszPathInfo:"");
	trace2("          Method = %s\n", lphrb->lpszMethod?lphrb->lpszMethod:"");
	trace2("          QueryString = %s\n", lphrb->lpszQueryString?lphrb->lpszQueryString:"");
	trace2("          PathTranslated = %s\n", lphrb->lpszPathTranslated?lphrb->lpszPathTranslated:"");
	trace2("          Cookie = %s\n", lphrb->lpszCookie?lphrb->lpszCookie:"");
	trace2("          ContentType = %s\n", lphrb->lpszContentType?lphrb->lpszContentType:"");
	trace2("          UserInfo = %s\n", lphrb->lpszUserInfo?lphrb->lpszUserInfo:"");
	trace2("          VirtualHost = %s\n", lphrb->lpszVirtualHost?lphrb->lpszVirtualHost:"");
	trace2("          VirtualHostRoot = %s\n", lphrb->lpszVirtualHostRoot?lphrb->lpszVirtualHostRoot:"");
	
	/* Set up client input */
	if ((retval = setup_client_block(r, REQUEST_CHUNKED_ERROR)))
		return retval;
	
	if (should_client_block(r)) 
	{
		long to_read = atol(table_get(e, "CONTENT_LENGTH"));
		long read, total_read;
		void *cur_read_pos = 0;

		/*		   
		if (to_read > HT_MAX_CONTENT_LENGTH) 
			return HTTP_REQUEST_ENTITY_TOO_LARGE;
		*/

		/* make room for the data */
		lphrb->lpData = pcalloc(r->pool, 1 + to_read);
		if (!lphrb->lpData)
			return SERVER_ERROR;
	
		total_read = 0;
		cur_read_pos = lphrb->lpData;
		
		/* read until we get all that we ask for */
		while (total_read < to_read)
		{
			if ((read = get_client_block(r, cur_read_pos, to_read)) < 0) 
				return SERVER_ERROR;
			total_read += read;
			/*((char *)cur_read_pos) += read; */
			cur_read_pos = (void*)((char*)cur_read_pos + read);
			if (!read) 
				break;
		}
		
		lphrb->cbAvailable = lphrb->cbTotalBytes = total_read;
		((char *)lphrb->lpData)[total_read] = '\0';
	}
	else 
	{
		lphrb->cbTotalBytes = 0;
		lphrb->cbAvailable = 0;
	}
	
	/* Set up the callbacks   */
	lphrb->get_server_var = &get_server_var;
	lphrb->write_client = &write_client;
	lphrb->read_client = &read_client;
	lphrb->server_function = &server_function;
	
	/* All right... send the request*/
	retval = casp_req(lphrb, &luStatus);

	/* set http status code */
	if (cid->custom_response)
	{
		r->status = HTTP_OK;
		ap_die(cid->custom_response, r);
	}
	else
	{
		r->status = luStatus;
	}
	
	switch(retval) 
	{
		case HT_STATUS_SUCCESS:
			if (cid->status) 
			{   
				/* We have a special status to return  */
				return cid->status;
			}
			return OK;
		case HT_STATUS_ERROR:
		default:
			return SERVER_ERROR;
	}
	
}
#ifdef WIN32
#pragma optimize("",on)
#endif

/* functions to return the uuencoded authentication */
/* string sent from the browser						*/
static char *get_auth_user(request_rec *r)
{
	const char *auth_line = ap_table_get(r->headers_in,
                                  r->proxyreq ? "Proxy-Authorization"
                                              : "Authorization");
	
	if (auth_line && !strcasecmp(ap_getword(r->pool, &auth_line, ' '), "Basic")) 
	{
		const char *t = ap_uudecode(r->pool, auth_line);
		return ap_getword_nulls (r->connection->pool, &t, ':');
    }
	else return NULL;
}

static char *get_auth_pw(request_rec *r)
{
	const char *auth_line = ap_table_get(r->headers_in,
                                  r->proxyreq ? "Proxy-Authorization"
                                              : "Authorization");
	
	if (auth_line && !strcasecmp(ap_getword(r->pool, &auth_line, ' '), "Basic")) 
	{
		const char *t = ap_uudecode(r->pool, auth_line);
		ap_getword_nulls (r->connection->pool, &t, ':');
		return (char *)t;
    }
	else return NULL;
}

static char *get_auth_type(request_rec *r)
{
	const char *auth_line = ap_table_get(r->headers_in,
                                  r->proxyreq ? "Proxy-Authorization"
                                              : "Authorization");
	
	if (auth_line)
		return ap_getword(r->pool, &auth_line, ' ');
	return NULL;
}


/* stuct used to locate server variables */
typedef struct _HTTP_HEADERS
{
	array_header a;
#ifdef MAKE_TABLE_PROFILE
	void *creator;
#endif
} HTTP_HEADERS;

/* called by the asp engine to retrieve server variables */
int STDCALL get_server_var(HCONN hConn, char *lpszVariableName,
						   void *lpvBuffer, unsigned long *lpdwSizeofBuffer) 
{
	request_rec *r = ((casp_cid *)hConn)->r;
	table *e = r->subprocess_env;
	HTTP_HEADERS *hhead = 0;
	char *result = 0;
	table_entry *elts = 0;
	int i;
	
	/* Mostly, we just grab it from the environment, but there are
	   a couple of special cases 
	*/
	if (!strcmp(lpszVariableName, "SERVER_PORT_SECURE")) 
	{
#ifndef EAPI
		/* Apache doesn't support secure requests inherently, so
		   we have no way of knowing. We'll be conservative, and say
		   all requests are insecure.  
		*/
		result = "0";
#else	
		char *method = ap_http_method(r);
		if (method && strcmp(method, "https") == 0)
			result = "1";
		else
			result = "0";
#endif /* EAPI */
	}
	else if (!strcmp(lpszVariableName, "ALL_HTTP")) 
	{
		hhead = (HTTP_HEADERS*)r->subprocess_env;
		elts = (table_entry*)hhead->a.elts;
		((char*)lpvBuffer)[0] = 0;
		for (i=0; i<hhead->a.nelts; i++)
		{
			if (!strncmp(elts[i].key, "HTTP_", 5))
			{
				strcat((char*)lpvBuffer, elts[i].key);
				strcat((char*)lpvBuffer, ":");
				strcat((char*)lpvBuffer, elts[i].val);
				strcat((char*)lpvBuffer, "\n");
			}
		}
		*lpdwSizeofBuffer = strlen((char*)lpvBuffer);
		return HT_STATUS_SUCCESS;
	}
	else if (!strcmp(lpszVariableName, "AUTH_PASSWORD"))
	{
		if ((result = get_auth_pw(r)) == NULL)
			result = 0;
	}
	else if (!strcmp(lpszVariableName, "AUTH_USER")) 
	{
		if ((result = get_auth_user(r)) == NULL)
			result = 0;
	}
	else if (!strcmp(lpszVariableName, "AUTH_TYPE")) 
	{
		if ((result = get_auth_type(r)) == NULL)
			result = 0;
	}
	else if (!strcmp(lpszVariableName, "LOGON_USER")) 
	{
		result = (char *)table_get(e, "REMOTE_USER");
	}
	else if (!strcmp(lpszVariableName, "ALL_RAW")) 
	{
		hhead = (HTTP_HEADERS*)r->headers_in;
		elts = (table_entry*)hhead->a.elts;
		((char*)lpvBuffer)[0] = 0;
		for (i=0; i<hhead->a.nelts; i++)
		{
			strcat((char*)lpvBuffer, elts[i].key);
			strcat((char*)lpvBuffer, ": ");
			strcat((char*)lpvBuffer, elts[i].val);
			strcat((char*)lpvBuffer, "\n");
		}
		*lpdwSizeofBuffer = strlen((char*)lpvBuffer);
		return HT_STATUS_SUCCESS;
	}
	else if (!strcmp(lpszVariableName, "URL")) 
	{
		result = r->uri;
	}
	else 
	{
		result = (char *)table_get(e, lpszVariableName);
	}
	
	if (result) 
	{
		unsigned long len = strlen(result);
		if (len > *lpdwSizeofBuffer) 
		{
			*lpdwSizeofBuffer = len;
			return HT_STATUS_ERROR;
		}
		strncpy(lpvBuffer, result, *lpdwSizeofBuffer);
		*lpdwSizeofBuffer = len;
		return HT_STATUS_SUCCESS;
	}
	*lpdwSizeofBuffer = 0;
	
	/* Didn't find it */
	return HT_STATUS_ERROR;
}

/* called by the asp engine to write data to client */
int STDCALL write_client(HCONN hConn, void *Buffer, unsigned long *lpwdwBytes) 
{
	request_rec *r = ((casp_cid *)hConn)->r;
	int writ;

	if (((casp_cid *)hConn)->custom_response)
	{
		return HT_STATUS_ERROR;
	}

	if ((writ = rwrite(Buffer, *lpwdwBytes, r)) == EOF) 
	{
		return HT_STATUS_ERROR;
	}
	rflush(r);
	
	*lpwdwBytes = writ;

	return HT_STATUS_SUCCESS;
}

/* called by the asp engine to read data from client */
int STDCALL read_client(HCONN hConn, void *lpvBuffer, unsigned long *lpdwSize) 
{
	/* Doesn't need to do anything; we've read all the data already */
	return HT_STATUS_SUCCESS;
}

/* called by the asp engine to send response header and log data */
int STDCALL server_function(HCONN hConn, unsigned long dwHSERequest,
							char *lpszBuffer, unsigned long *lpdwSize, char *lpszExtra) 
{
	casp_cid *cid = (casp_cid *)hConn;
	request_rec *subreq = 0; 
	request_rec *r = cid->r;
	char *data = 0;

	switch (dwHSERequest) 
	{
	case HT_SFUNC_LOGDATA:
		/* Log an error */
		if (!lpszBuffer)
		{ 
			log_error("chiliasp: engine sent invalid log data.", r->server);
			return HT_STATUS_ERROR;
		}
		
		log_error(lpszBuffer, r->server);
		
		return HT_STATUS_SUCCESS;

	case HT_SFUNC_RESPONSEHEADERS:
		/* this function may only be called once with HT_SFUNC_RESPONSEHEADERS */
		
		if (r->status_line)
		{
			/*printf("chiliasp: HT_SFUNC_RESPONSEHEADERS sent more than once.\n");*/
			return HT_STATUS_SUCCESS;
		}
	
		r->status_line = lpszBuffer ? lpszBuffer : pstrdup(r->pool, "200 OK");
		sscanf(r->status_line, "%d", &r->status);

		/* does this response have a custom response */
		if (r->status != 401 && r->status != 403 &&
			ap_response_code_string(r, ap_index_of_response(r->status)))
		{
			/* this status has a custom response, let apache handle it */
			((casp_cid *)hConn)->custom_response = r->status;
			r->status = HTTP_OK;
			r->status_line = NULL;
			return HT_STATUS_ERROR;
		}

		/* Fill in the HTTP headers, and the rest of it. Ick.
		   lpdwDataType contains a string that has headers (in MIME
		   format), a blank like, then (possibly) data. We need
		   to parse it.
		  
		   Easy case first:
		*/

		if (!lpszExtra) 
		{
			send_http_header(r);
			return HT_STATUS_SUCCESS;
		}
		
		/* Make a copy - don't disturb the original */
		data = pstrdup(r->pool, (char *)lpszExtra);
		
		/* We *should* break before this while loop ends */
		while (*data) 
		{
			char *lf = strchr(data, '\n');
			char *value = 0; 
			int p;
			
			if (lf)
				*lf = '\0';

			p = strlen(data);
			if (p > 0 && data[p-1] == '\r') data[p-1] = '\0';
			
			/* End of headers */
			if (*data == '\0') 
			{
				if (lf)
					data = lf + 1;
				break;
			}
			
			if (!(value = strchr(data, ':'))) 
			{                         
				error("chiliasp: engine sent invalid headers", r->server);
				return HT_STATUS_ERROR;
			}
			
			*value++ = '\0';
			while (*value && isspace(*value)) ++value;
			
			/* Check all the special-case headers. Similar to what
			   scan_script_header() does (see that function for
			   more detail)
			*/
			if (!strcasecmp(data, "Content-Type")) 
			{
				char *tmp;
				/* Nuke trailing whitespace */
				
				char *endp = value + strlen(value) - 1;
				while (endp > value && isspace(*endp)) *endp-- = '\0';
				
				tmp = pstrdup (r->pool, value);
				str_tolower(tmp);
				r->content_type = tmp;
			}
			else if (!strcasecmp(data, "Content-Length")) {
				table_set(r->headers_out, data, value);
			}
			else if (!strcasecmp(data, "Transfer-Encoding")) {
				table_set(r->headers_out, data, value);
			}
			else if (!strcasecmp(data, "Set-Cookie")) {
				table_add(r->err_headers_out, data, value);
			}
			else {
				table_merge(r->err_headers_out, data, value);
			}
			
			/* Reset data */
			if (!lf) {
				data += p;
				break;
			}
			data = lf + 1;
		}
		
		/* All the headers should be set now */
		send_http_header(r);
		
		/* Any data left should now be sent directly */
		rputs(data, r);
		
		return HT_STATUS_SUCCESS;

	default:
		return HT_STATUS_ERROR;
	}
}





/* per-server configuration data */
typedef struct {
	char *identifier;
	char *docroot;
	char *user;
	char *group;
	char *user_slash_group;
	int allowed;
} VSERV_INFO;


/* does this uri match any of our asp applications? */
char *try_app_list(request_rec *r, const char *hostname)
{
	char *location = NULL;
	int l = search_applications(r->uri, hostname, &location);
	
	if (l > 0)
	{
		trace3("chiliasp: uri matched chars = %i\n", l);
		trace3("          uri [%s], loc [%s]\n", r->uri, location);
    	
		return pstrcat(r->pool, location, r->uri + l, NULL);
	}
	
	trace3("chiliasp: uri *didn't* match [%s]\n", r->uri);
	return NULL;
}
		
/* 
	does virtual translations for asp apps not 
	defined as a virtual root 
*/
static int casp_translate(request_rec *r)
{
	char *ret = 0;

	/* we are not going to translate anything */
	/* but the real host					  */
	VSERV_INFO *pvsi = (VSERV_INFO*)get_module_config(
							r->server->module_config, &casp2_module);

	if (!pvsi || !pvsi->identifier || 
		(pvsi->identifier[0] != '/' && pvsi->identifier[1] != 0))
		return DECLINED;
	
	/* don't translate anything that doesn't */
	/* begin with a front slash				 */
	if (r->uri[0] != '/') 
		return DECLINED;
	
	/* don't attempt to translate the doc root */ 
	if (r->uri[0] == '/' && r->uri[1] == '\0') 
		return DECLINED;

	/* check if it matches one of our asp apps */
    if ((ret = try_app_list(r, pvsi->identifier)) != NULL) 
    {
        r->filename = ret;
		
		trace3("chiliasp: uri resolved to [%s]\n", r->filename);
        
		return OK;
    }
    
    return DECLINED;
}


/* sets the directory for the chiliasp bin directory */
static const char *casp_lib_directory(cmd_parms *cmd, void *dummy, char *arg)
{
	/* make sure libs have been loaded */
	init_casp2_module();
	
	/* store this path */
	strncpy(CaspLibDir, arg, 1023);
		
    return NULL;
}

/* returns the servers id from the config data */
const char *get_server_id(request_rec *r)
{
	VSERV_INFO *pvsi = (VSERV_INFO*)get_module_config(
							r->server->module_config, &casp2_module);

	/* the virtual host must have the 'ServerName' */
	/* directive in order to be used with chiliasp */
	if (!pvsi || !pvsi->identifier)
		return HT_REALHOST_ID;

	return pvsi->identifier;
}

/* returns the servers docroot from the config data */
const char *get_server_docroot(request_rec *r)
{
	VSERV_INFO *pvsi = (VSERV_INFO*)get_module_config(
							r->server->module_config, &casp2_module);

	if (!pvsi)
		return NULL;

	if (pvsi->allowed == 0)
		return NULL;

	return pvsi->docroot;
}

/* returns the servers user/group string from the config data */
const char *get_usergroup_string(request_rec *r)
{
	static char empty[] = "";
	VSERV_INFO *pvsi = (VSERV_INFO*)get_module_config(
							r->server->module_config, &casp2_module);

	if (!pvsi)
		return NULL;

	/* the user_slash_group string needs to be */
	/* if it is zero						   */
	if (!pvsi->user_slash_group)
    {
		char temp[1024] = "";
		if (!pvsi->user)
		{
#ifdef WIN32
			pvsi->user = empty;
#else
			struct passwd *psd = getpwuid(r->server->server_uid);
			pvsi->user = strdup(psd->pw_name ? psd->pw_name : empty);
#endif
		}
		strcat(temp, pvsi->user);
		strcat(temp, "/");
		if (!pvsi->group)
		{
#ifdef WIN32
			pvsi->group = empty;
#else
			struct group *pgp = getgrgid(r->server->server_gid);
			pvsi->group = strdup(pgp->gr_name ? pgp->gr_name : empty);
#endif
		}
		strcat(temp, pvsi->group);
	
		pvsi->user_slash_group = strdup(temp);
		
		trace2("chiliasp: user/group=%s\n", pvsi->user_slash_group);
    }
	return pvsi->user_slash_group;
}

/* 
	this is the notification of our server's 
	document roots. there will be a document 
	root for the main server and each virtual
	server. we add each document root to the 
	list of asp applications				
*/
static const char *doc_root(cmd_parms *cmd, void *dummy, char *arg)
{
	VSERV_INFO *pvsi = (VSERV_INFO*)get_module_config(
							cmd->server->module_config, &casp2_module);

	/* make sure libs have been loaded */
	init_casp2_module();
	
	if (!pvsi || !arg)
		return "chiliasp: error reading 'DocumentRoot' directives.";

#ifdef VER_13
	pvsi->docroot = os_canonical_filename(cmd->pool, arg);
#else
	pvsi->docroot = strdup(arg);
#endif /* VER_13 */
	
	trace2("chiliasp: docroot found [%s]\n", pvsi->docroot);

	return DECLINE_CMD;
}

static const char *server_name(cmd_parms *cmd, void *dummy, char *arg)
{
	VSERV_INFO *pvsi =
	  (VSERV_INFO*)get_module_config(cmd->server->module_config,
					 &casp2_module);

	/* make sure libs have been loaded */
	init_casp2_module();
	
	if (!pvsi || !arg)
	{
		error("chiliasp: error reading 'ServerName' directives.\n");

		return "chiliasp: error reading 'ServerName' directives.";
	}

	if (cmd->server->is_virtual)
	{
		pvsi->identifier = strdup(arg);
		trace1("chiliasp: virtual host found [%s]\n", pvsi->identifier);
	}
	else
	{
		pvsi->identifier = strdup(HT_REALHOST_ID);
		if (real_host_name == NULL)
			real_host_name = strdup(arg);
		trace1("chiliasp: real host found [%s]\n", pvsi->identifier);
	}

#ifdef WIN32
	// WINNT uses backslashes in all cases here.
	pvsi->identifier[0] = '\\';
#endif
	
	return DECLINE_CMD;
}

static const char *alias_found(cmd_parms *cmd, void *dummy, char *arg1, char *arg2)
{
	VSERV_INFO *pvsi = (VSERV_INFO*)get_module_config(
							cmd->server->module_config, &casp2_module);

	/* make sure libs have been loaded */
	init_casp2_module();
	
	if (!arg1 || !arg2 || !pvsi)
		return "chiliasp: error reading 'Alias' directives.";
		
	/* remove trailing backslashes */
	arg1 = strdup(arg1);
	if (arg1[strlen(arg1) - 1] == '/')
		arg1[strlen(arg1) - 1] = 0;

	arg2 = strdup(arg2);
	if (arg2[strlen(arg2) - 1] == '/')
		arg2[strlen(arg2) - 1] = 0;

	trace2("chiliasp: alias found %s [%s=%s]\n", 
		pvsi->identifier?pvsi->identifier:HT_REALHOST_ID, arg1, arg2);

	/* add it as an asp application */
	add_application(arg1, arg2, pvsi->identifier, 1);

	return DECLINE_CMD;
}

static const char *user_found(cmd_parms *cmd, void *dummy, char *arg)
{
	VSERV_INFO *pvsi = (VSERV_INFO*)get_module_config(
							cmd->server->module_config, &casp2_module);

	/* make sure libs have been loaded */
	init_casp2_module();
	
	if (!pvsi || !arg)
		return "chiliasp: error reading 'User' directives.";

	pvsi->user = strdup(arg);

	trace2("chiliasp: user found [%s]\n", pvsi->user);

	return DECLINE_CMD;
}

static const char *group_found(cmd_parms *cmd, void *dummy, char *arg)
{
	VSERV_INFO *pvsi = (VSERV_INFO*)get_module_config(
							cmd->server->module_config, &casp2_module);

	/* make sure libs have been loaded */
	init_casp2_module();
	
	if (!pvsi || !arg)
		return "chiliasp: error reading 'Group' directives.";

	pvsi->group = strdup(arg);

	trace2("chiliasp: group found [%s]\n", pvsi->group);

	return DECLINE_CMD;
}



static void casp_initializer( server_rec *Server, pool *Pool )
{
	/* make sure lib dir was set */
	if (strlen(CaspLibDir) == 0)
	{
		error("chiliasp: library directory not set yet.\n");
		delete_applications();
		exit(FAIL);
	}
	
	/* initialize module */
	if (initialize_casp_libs(CaspLibDir, g_DebugFlag) != OK)
	{
		error("chiliasp: failed to initialize web server module.\n");
		delete_applications();
		exit(FAIL);
	}
}



/* 
	this function is called for all virtual servers 
 	and the main server. we use it to allocate the  
	config stucture for each server.				   
*/
static void *new_server(pool *p, server_rec *pserver)
{
	VSERV_INFO *pvsi = (VSERV_INFO*)palloc(p, sizeof(VSERV_INFO));
	
	/* make sure libs have been loaded */
	init_casp2_module();
	
	if (!pvsi)
	{
		error("chiliasp: error allocating per-server data.\n");
		exit(1);
	}

	/* initialize allocation */
	pvsi->identifier = "/";
	pvsi->docroot = 0;
	pvsi->user = 0;
	pvsi->group = 0;
	pvsi->user_slash_group = 0;
	pvsi->allowed = 1;
	return pvsi;
}

/* make sure we have what we need */
static void *merge_server(pool *p, void *basev, void *addv)
{
	VSERV_INFO *pvsib = (VSERV_INFO*)basev;
	VSERV_INFO *pvsia = (VSERV_INFO*)addv;

	/* make sure libs have been loaded */
	init_casp2_module();

	/* initialize module */
	casp_initializer( NULL, NULL );
	
	if (!pvsib || !pvsia)
	{
		error("chiliasp: error merging per-server data.\n");
		exit(FAIL);
	}

	/* check virtual */
	if (!pvsia->identifier)
	{
		error("chiliasp: 'ServerName' directive missing for virtual host.\n");
		exit(FAIL);
	}

    if (real_host_name && !strcasecmp(real_host_name, pvsia->identifier))
    {
    	trace1("chiliasp: virtual host %s is real host\n", pvsia->identifier);
        pvsia->allowed = is_virtual_host_allowed(pvsia->identifier);
        trace1("chiliasp: virtual host %s %s\n",
                pvsia->identifier, (pvsia->allowed)?"allowed":"not allowed");
        free(pvsia->identifier);
        pvsia->identifier = strdup(HT_REALHOST_ID);
    }      
    else
    {
        /* check if virtual host is allowed */
        pvsia->allowed = is_virtual_host_allowed(pvsia->identifier);
        trace1("chiliasp: virtual host %s %s\n",
                pvsia->identifier, (pvsia->allowed)?"allowed":"not allowed");
    }

	/* check user/group stuff */
	if (!pvsia->user && pvsib->user)
		pvsia->user = pvsib->user;
	if (!pvsia->group && pvsib->group)
		pvsia->group = pvsib->group;

	return addv;
}

static const char *set_casp_debugging(cmd_parms *cmd, void *dummy, char *arg)
{
	/* turn debugging on */
	if (arg)
	{
		sscanf(arg, "%d", &g_DebugFlag);
	}
	
	if (!arg || g_DebugFlag < 0 || g_DebugFlag > 5)
	{
		error("chiliasp: CaspDebug ranges from 0 to 5.\n");
	}
	
	return NULL;
}

/* 	
	the command table - our module will be notified 
	when any of the following commands are found    
 	in the web server's configuration files     
 	
 	the format for the command table is:
 	{COMMAND, CALLBACK, EXTRADATA, OVERRIDES, ARGS, ERRMSG}
 	
 	where:
 	
 	COMMAND		= the command to look for in config files
 	CALLBACK	= function that Apache will call in this module
 	EXTRADATA	= not used
 	OVERRIDES	= where the command is permitted
 	ARGS		= format of arguments expected
 	ERRMSG		= error message displayed when command not found
*/    
												
#ifdef WIN32
static const 
#endif
command_rec cnfg_cmds[] = 
{
	{"CaspLib", casp_lib_directory, NULL, OR_FILEINFO, TAKE1,
	    "chiliasp: location of chiliasp lib directory."},
	{"CaspDebug", set_casp_debugging, NULL, RSRC_CONF, TAKE1,
	    "chiliasp: debug flag for chiliasp module."},
	{"DocumentRoot", doc_root, NULL, RSRC_CONF, TAKE1,
	    "chiliasp: server's document root."},
	{"ServerName", server_name, NULL, RSRC_CONF, TAKE1,
	    "chiliasp: server's hostname."},
	{"Alias", alias_found, NULL, RSRC_CONF, TAKE2,
	    "chiliasp: Alias."},
	{"User", user_found, NULL, RSRC_CONF, TAKE1,
	    "chiliasp: User."},
	{"Group", group_found, NULL, RSRC_CONF, TAKE1,
	    "chiliasp: Group."},	
	{NULL}
};

handler_rec casp_handlers[] = 
{
	{"chiliasp", casp_handler},
	{NULL}
};

module MODULE_VAR_EXPORT casp2_module = 
{
	STANDARD_MODULE_STUFF,
	casp_initializer,			/* initializer */
	NULL,						/* create per-dir config */
	NULL,						/* merge per-dir config */
	new_server,					/* server config */
	merge_server,				/* merge server config */
	cnfg_cmds,					/* command table */
	casp_handlers, 				/* handlers */
#ifdef WIN32
	NULL,						/* filename translation */
#else
	casp_translate,				/* filename translation */
#endif
	NULL,						/* check_user_id */
	NULL,						/* check auth */
	NULL,						/* check access */
	NULL,						/* type_checker */
	NULL,						/* fixups */
	NULL,						/* logger */
	NULL						/* header parser */
#ifdef WIN32
		,
	NULL,						/* child init */
	exit_handler,				/* child exit */
	NULL						/* post-read request */
#endif
};




